package clime.messadmin.filter;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import clime.messadmin.core.Constants;
import clime.messadmin.model.Server;
import clime.messadmin.providers.ProviderUtils;
import clime.messadmin.providers.spi.RequestExceptionProvider;
import clime.messadmin.providers.spi.RequestLifeCycleProvider;

/**
 * Servlet Filter which logs request stats and injects a message in the output html stream
 * @author C&eacute;drik LIME
 */
public class MessAdminFilter implements Filter {
	private static boolean DEBUG = false;
	private FilterConfig config;

	private void log(final String message) {
		if (DEBUG) {
			config.getServletContext().log(message);
		}
	}
	private void log(final String message, final Throwable t) {
		if (DEBUG) {
			config.getServletContext().log(message, t);
		}
	}

	/**
	 * 
	 */
	public MessAdminFilter() {
		super();
	}

	/**
	 * {@inheritDoc}
	 */
	public void init(final FilterConfig filterConfig) throws ServletException {
		this.config = filterConfig;
	}

	/**
	 * {@inheritDoc}
	 */
	public void destroy() {
		config = null;
	}

	/**
	 * {@inheritDoc}
	 */
	public void doFilter(final ServletRequest request, final ServletResponse response,
			final FilterChain chain) throws IOException, ServletException {
		if ( ! (request instanceof HttpServletRequest && response instanceof HttpServletResponse)) {
			// not http, don't do anything
			chain.doFilter(request, response);
			return;
		}
//		final long before0 = System.nanoTime();
		final HttpServletRequest httpRequest = (HttpServletRequest) request;
		final HttpSession session = httpRequest.getSession(false);
		final MessAdminRequestWrapper wrappedRequest = new MessAdminRequestWrapper(httpRequest);
		final MessAdminResponseWrapper wrappedResponse = new MessAdminResponseWrapper((HttpServletResponse) response);

		//assert session == null || session.getServletContext() == config.getServletContext();

		// See if there is a message to inject in this request/response, and if so record it
		boolean messageFromSession = fetchDisplayMessage(session, config.getServletContext(), wrappedResponse);

		MessAdminThreadLocal.start();

		// Sniff request infos for future usage
		Server.getInstance().requestInitialized(httpRequest, config.getServletContext());

		try {
			// pre-request plugin
			List requestProviders = ProviderUtils.getProviders(RequestLifeCycleProvider.class);
			Iterator iter = requestProviders.iterator();
			while (iter.hasNext()) {
				RequestLifeCycleProvider lc = (RequestLifeCycleProvider) iter.next();
				try {
					lc.requestInitialized(wrappedRequest, wrappedResponse, config.getServletContext());
				} catch (RuntimeException rte) {
				}
			}

//			final long before1 = System.nanoTime();
			chain.doFilter(wrappedRequest, wrappedResponse);
			wrappedResponse.flushBuffer();
//			final long after0 = System.nanoTime();

			MessAdminThreadLocal.stop();

			wrappedResponse.finish();

			HttpSession sessionAfter = httpRequest.getSession(false);
			// If injected message came from session and was not injected, put it back in session
			if (messageFromSession) {
				restoreDisplayMessage(session, wrappedResponse);
			} else if (sessionAfter != null && wrappedResponse.isMessageInjected()) {
				// message is application-level and has been injected: record timestamp
				try {
					sessionAfter.setAttribute(Constants.GLOBAL_MESSAGE_TIMESTAMP_KEY, new Long(System.currentTimeMillis()));
				} catch (IllegalStateException ise) {
					// session is invalidated: do nothing
				}
			}

			if (session == null && sessionAfter != null) {
				// Session was created
				Server.getInstance().requestInitialized(httpRequest, config.getServletContext());
			}
			Server.getInstance().requestDestroyed(wrappedRequest, wrappedResponse, config.getServletContext());

			// post-request plugin
			requestProviders = new ArrayList(requestProviders);
			Collections.reverse(requestProviders);
			iter = requestProviders.iterator();
			while (iter.hasNext()) {
				RequestLifeCycleProvider lc = (RequestLifeCycleProvider) iter.next();
				try {
					lc.requestDestroyed(wrappedRequest, wrappedResponse, config.getServletContext());
				} catch (RuntimeException rte) {
				}
			}

//			final long after1 = System.nanoTime();

//			long before = before1 - before0;
//			long after = after1 - after0;
//			System.out.println("Before: " + before/1000000.0 + " ms\tAfter: " + after/1000000.0 + " ms\tTotal: " + (before+after)/1000000.0 + " ms.");
		} catch (IOException ioe) {
			reportException(wrappedRequest, wrappedResponse, ioe);
			throw ioe;
		} catch (ServletException se) {
			reportException(wrappedRequest, wrappedResponse, se);
			throw se;
		} catch (RuntimeException rte) {
			reportException(wrappedRequest, wrappedResponse, rte);
			throw rte;
		} finally {
			MessAdminThreadLocal.remove();
		}
	}

	private void reportException(MessAdminRequestWrapper request, MessAdminResponseWrapper response, Exception e) {
		MessAdminThreadLocal.stop();
		Server.getInstance().requestException(e, request, response, config.getServletContext());
		Iterator iter = ProviderUtils.getProviders(RequestExceptionProvider.class).iterator();
		while (iter.hasNext()) {
			RequestExceptionProvider lc = (RequestExceptionProvider) iter.next();
			try {
				lc.requestException(e, request, response, config.getServletContext());
			} catch (RuntimeException rte) {
			}
		}
	}

	/**
	 * Get the session-specific message to display, or the application-specific one if no session-message exist.
	 * @param wrappedResponse
	 * @param session
	 * @return <code>true</code> if message was from user session, <code>false</code>otherwise
	 */
	private boolean fetchDisplayMessage(final HttpSession session, final ServletContext servletContext, final MessAdminResponseWrapper wrappedResponse) {
		String message = null;
		boolean fromSession = false;
		// try session-level message
		if (session != null) {
			try {
				message = (String) session.getAttribute(Constants.SESSION_MESSAGE_KEY);
				if (message != null) {
					// display session messages only once
					log("Removing session message for session id " + session.getId());//$NON-NLS-1$
					session.removeAttribute(Constants.SESSION_MESSAGE_KEY);
					fromSession = true;
				}
			} catch (IllegalStateException ise) {
				// invalidated session: don't do anything
			}
		}
		// try application-level message
		if (message == null) {
			message = (String) servletContext.getAttribute(Constants.GLOBAL_MESSAGE_KEY);
			// only display application-level message if not displayed "too" recently
			if (message != null && session != null) {
				Long lastTimeStamp = (Long) session.getAttribute(Constants.GLOBAL_MESSAGE_TIMESTAMP_KEY);
				if (lastTimeStamp != null) {
					if (System.currentTimeMillis() - lastTimeStamp.longValue() < Constants.GLOBAL_MESSAGE_DELTA_TIME_MIN) {
						message = null;
					}
				}
			}
		}
		//TODO try server-level message
		if (message != null) {
			wrappedResponse.setInjectedMessageHTML(message);
		}
		return fromSession;
	}

	private void restoreDisplayMessage(final HttpSession session, final MessAdminResponseWrapper wrappedResponse) {
		String message = wrappedResponse.getInjectedMessageHTML();
		if (session != null && message != null) {
			try {
				String newMessage = (String) session.getAttribute(Constants.SESSION_MESSAGE_KEY);
				if (newMessage == null) {
					// only put back message if there isn't a new one
					log("Putting back session message for session id " + session.getId());//$NON-NLS-1$
					session.setAttribute(Constants.SESSION_MESSAGE_KEY, message);
				}
			} catch (IllegalStateException ise) {
				// invalidated session: don't do anything
			}
		}
	}
}
