/**
 * 
 */
package clime.messadmin.providers.lifecycle;

import java.beans.Introspector;
import java.lang.reflect.Method;
import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Enumeration;

import javax.servlet.ServletContext;
import javax.servlet.ServletContextEvent;

import clime.messadmin.providers.spi.ApplicationLifeCycleProvider;

/**
 * Takes care of deregistering (some of) the evil Singletons
 * when the app shuts down, thereby avoiding (well, trying to avoid)
 * OOM (java.lang.OutOfMemoryError) on hot restart...
 * 
 * @see <a href="http://opensource2.atlassian.com/confluence/spring/pages/viewpage.action?pageId=2669">Memory leaks where the classloader cannot be garbage collected</a>
 * @author C&eacute;drik LIME
 */
public class EvilSingletonsUnregisterer implements ApplicationLifeCycleProvider {

	/**
	 * 
	 */
	public EvilSingletonsUnregisterer() {
		super();
	}

	/**
	 * {@inheritDoc}
	 */
	public int getPriority() {
		return Integer.MAX_VALUE;
	}

	/**
	 * {@inheritDoc}
	 */
	public void contextDestroyed(ServletContext servletContext) {
		// ThreadLocals - With Great Power, comes Great Responsibility
		// Can't do anything here. You need to manually set all your ThreadLocals to null yourself (myThreadLocal.remove() or myThreadLocal.set(null))...

		final ClassLoader thisClassLoader = Thread.currentThread().getContextClassLoader();//this.getClass().getClassLoader();

		// java.beans.Introspector - Slightly less Evil singleton
		// Flushes the cache of classes
		// Note this is going to clear ALL the classes, regardless of what ClassLoader or application they came from, but it's all that's available.
		Introspector.flushCaches();

		// Jakarta Commons IO >= 1.3
		// org.apache.commons.io.FileCleaner.exitWhenFinished();
		try {
			Class fileCleanerClass = thisClassLoader.loadClass("org.apache.commons.io.FileCleaner");//$NON-NLS-1$
			if (fileCleanerClass != null && fileCleanerClass.getClassLoader() == thisClassLoader) {
				// If the given classloader is the same as the classloader of FileCleaner, this means that the lib
				// has been deployed inside the war, so the thread should be stopped and all resources released.
				// If the classloader is different, then the thread is owned by the container, so don't stop it.
				Method ewfMethod = fileCleanerClass.getMethod("exitWhenFinished", null);//$NON-NLS-1$
				ewfMethod.invoke(null, null);
			}
		} catch (Throwable t) {
			// ignore
		}

		// Jakarta Commons Logging <= 1.0.4
		// org.apache.commons.logging.LogFactory.release(Thread.currentThread().getContextClassLoader());
		try {
			ClassLoader loader = thisClassLoader;
			while (loader != null) {
				Class logFactoryClass = loader.loadClass("org.apache.commons.logging.LogFactory");//$NON-NLS-1$
				Method releaseMethod = logFactoryClass.getMethod("release", new Class[] {ClassLoader.class});//$NON-NLS-1$
				releaseMethod.invoke(null, new Object[] {thisClassLoader});
				loader = logFactoryClass.getClassLoader().getParent();
			}
		} catch (Throwable t) {
			// ignore
		}

		// LOGBack
		// new ch.qos.logback.classic.selector.servlet.ContextDetachingSCL().contextDestroyed(ServletContextEvent sce)
		try {
			Class cdClass = thisClassLoader.loadClass("ch.qos.logback.classic.selector.servlet.ContextDetachingSCL");//$NON-NLS-1$
			Object instance = cdClass.newInstance();
			Method destroyMethod = cdClass.getMethod("contextDestroyed", new Class[] {ServletContextEvent.class});//$NON-NLS-1$
			destroyMethod.invoke(instance, new Object[] {new ServletContextEvent(servletContext)});
		} catch (Throwable t) {
			// ignore
		}

		// Apache Logging Log4J 1.x
		// org.apache.log4j.LogManager.shutdown();
		try {
			Class lmClass = thisClassLoader.loadClass("org.apache.log4j.LogManager");//$NON-NLS-1$
			Method shutdownMethod = lmClass.getMethod("shutdown", null);//$NON-NLS-1$
			shutdownMethod.invoke(null, null);
		} catch (Throwable t) {
			// ignore
		}

		// java.sql.DriverManager - Evil Singleton
		// Although registering the JDBC driver in your web app is a horrible, horrible thing to do (the container should always manage your connections), some apps do just that.
		// Unregister JDBC drivers during shutdown: remove any drivers that were loaded by the same classloader that loaded the web app.
		Enumeration drivers = DriverManager.getDrivers();
		while (drivers.hasMoreElements()) {
			Driver o = (Driver) drivers.nextElement();
			if (o.getClass().getClassLoader() == thisClassLoader){
				// Current driver 'o' is being deregistered
				try {
					DriverManager.deregisterDriver(o);
				} catch (SQLException sqle) {
					//throw new RuntimeException(sqle.getMessage());//new RuntimeException(sqle);
					System.err.println("Failed to cleanup DriverManager for webapp:");
					sqle.printStackTrace(System.err);
				}
			} else {
				// Driver 'o' wasn't loaded by this webapp, so no touching it
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public void contextInitialized(ServletContext servletContext) {
	}

}
