package clime.messadmin.utils;

import java.io.Serializable;

/**
 * Comparable Boolean for JDK <= 1.4. Useless for JDK 5.
 * Note that some static methods have been removed, as they where already defined in Boolean.
 * This class is not intended as a replacement for java.lang.Boolean!
 * 
 * (From the JDK documentation:)
 * The Boolean class wraps a value of the primitive type 
 * <code>boolean</code> in an object. An object of type 
 * <code>Boolean</code> contains a single field whose type is 
 * <code>boolean</code>. 
 * <p>
 * In addition, this class provides many methods for 
 * converting a <code>boolean</code> to a <code>String</code> and a 
 * <code>String</code> to a <code>boolean</code>, as well as other 
 * constants and methods useful when dealing with a 
 * <code>boolean</code>. 
 *
 * @author  Arthur van Hoff, C&eacute;drik LIME
 * @version 1.51, 05/11/04
 * @since   JDK1.0
 */
public final class ComparableBoolean implements Serializable, Comparable {//extends Boolean

    /** 
     * The <code>ComparableBoolean</code> object corresponding to the primitive 
     * value <code>true</code>. 
     */
    public static final ComparableBoolean TRUE = new ComparableBoolean(true);

    /** 
     * The <code>ComparableBoolean</code> object corresponding to the primitive 
     * value <code>false</code>. 
     */
    public static final ComparableBoolean FALSE = new ComparableBoolean(false);

    /**
     * The Class object representing the primitive type boolean.
     *
     * @since   JDK1.1
     */
    public static final Class TYPE = Boolean.TYPE;

    /**
     * The value of the Boolean.
     *
     * @serial
     */
    private final boolean value;

    /** use serialVersionUID from JDK 1.0.2 for interoperability */
    //private static final long serialVersionUID = -3665804199014368530L;

    /**
     * Allocates a <code>ComparableBoolean</code> object representing the 
     * <code>value</code> argument. 
     *
     * <p><b>Note: It is rarely appropriate to use this constructor.
     * Unless a <i>new</i> instance is required, the static factory
     * {@link #valueOf(boolean)} is generally a better choice. It is
     * likely to yield significantly better space and time performance.</b>
     * 
     * @param   value   the value of the <code>Boolean</code>.
     */
    private ComparableBoolean(boolean value) {
    this.value = value;
    }

    /**
     * Returns the value of this <tt>ComparableBoolean</tt> object as a boolean 
     * primitive.
     *
     * @return  the primitive <code>boolean</code> value of this object.
     */
    public boolean booleanValue() {
    return value;
    }

    /**
     * Returns a <tt>ComparableBoolean</tt> instance representing the specified
     * <tt>boolean</tt> value.  If the specified <tt>boolean</tt> value
     * is <tt>true</tt>, this method returns <tt>ComparableBoolean.TRUE</tt>;
     * if it is <tt>false</tt>, this method returns <tt>ComparableBoolean.FALSE</tt>.
     * If a new <tt>ComparableBoolean</tt> instance is not required, this method
     * should generally be used in preference to the constructor
     * {@link #ComparableBoolean(boolean)}, as this method is likely to yield
     * significantly better space and time performance.
     *
     * @param  b a boolean value.
     * @return a <tt>ComparableBoolean</tt> instance representing <tt>b</tt>.
     * @since  1.4
     */
    public static ComparableBoolean valueOf(boolean b) {
        return (b ? TRUE : FALSE);
    }

    /**
     * Returns a <tt>String</tt> object representing this ComparableBoolean's
     * value.  If this object represents the value <code>true</code>,
     * a string equal to <code>"true"</code> is returned. Otherwise, a
     * string equal to <code>"false"</code> is returned.
     *
     * @return  a string representation of this object. 
     */
    public String toString() {
    return value ? "true" : "false";//Boolean.toString(value);//$NON-NLS-1$//$NON-NLS-2$
    }

    /**
     * Returns a hash code for this <tt>ComparableBoolean</tt> object.
     *
     * @return  the integer <tt>1231</tt> if this object represents 
     * <tt>true</tt>; returns the integer <tt>1237</tt> if this 
     * object represents <tt>false</tt>. 
     */
    public int hashCode() {
    return value ? 1231 : 1237;//Boolean.valueOf(value).hashCode();
    }

    /**
     * Returns <code>true</code> if and only if the argument is not 
     * <code>null</code> and is a <code>ComparableBoolean</code> object that 
     * represents the same <code>boolean</code> value as this object. 
     *
     * @param   obj   the object to compare with.
     * @return  <code>true</code> if the Boolean objects represent the 
     *          same value; <code>false</code> otherwise.
     */
    public boolean equals(Object obj) {
    if (obj instanceof ComparableBoolean) {
        return value == ((ComparableBoolean)obj).booleanValue();
    } 
    if (obj instanceof Boolean) {
        return value == ((Boolean)obj).booleanValue();
    } 
    return false;
    }

    /**
     * Compares this <tt>ComparableBoolean</tt> instance with another.
     *
     * @param   o the <tt>ComparableBoolean</tt> or <tt>Boolean</tt> instance to be compared
     * @return  zero if this object represents the same boolean value as the
     *          argument; a positive value if this object represents true
     *          and the argument represents false; and a negative value if
     *          this object represents false and the argument represents true
     * @throws  NullPointerException if the argument is <tt>null</tt>
     * @see     Comparable
     * @since  1.5
     */
    public int compareTo(Object o) {
        if (o instanceof ComparableBoolean) {
            Boolean bool = ((ComparableBoolean)o).value ? Boolean.TRUE : Boolean.FALSE;//Boolean.valueOf(((ComparableBoolean)o).value);
            return compareTo(bool);
        } else if (o instanceof Boolean) {
            return (((Boolean)o).booleanValue() == value ? 0 : (value ? 1 : -1));
        }
        throw new ClassCastException(o.getClass().getName());
    }

}
